
# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####



# DESCRIPTION
# 
# Scales and snap-stretches one or more objects with one end fixed.
#
# It's main advantage is stretching with precision. I am used of working with simmilar commands a lot, during my architecture design proceses. Good for construction beams, openings, table-tops, carpets and so on. Besides that, it doesn't require object origin nad 3D cursor manipulation. It seems eficient.
# 
# INSTALATION
# 
# Two ways:
# 
# A. Paste the the .py file to text editor and run (ALT+P)
# B. Unzip and place .py file to addons_contrib. In User Preferences / Addons tab search under Testing / NP Cage Scale and check the box.
# 
# Now you have the operator in your system. If you press Save User Preferences, you will have it at your disposal every time you run Bl.
# 
# SHORTCUTS
# 
# After succesful instalation of the addon, or it's activation from the text editor, the NP Cage Scale operator, along with several others should be registered in your system. Enter User Preferences / Input and under 3DView / Object Mode add new keymap definition. Type object.np_cage_scale_xxx (xxx being the number of the version) and assign shortcut. I suggest asigning S only for the Object Mode because the addon doesn't work in other modes and it is only a prototype. This way the basic S command should be available in all other editing environments and you can copy it back if needed.
# 
# USAGE
# 
# Select one or more objects. 
# Run operator (spacebar search - NP Cage Scale, or keystroke if you assigned it)
# Pick-snap one of the points on the mode-cage.
# Scale or stretch with CTRL-snap :)
# 
# IMPORTANT PERFORMANCE NOTES
#
# All origins of objects participating are centered to their geometry after scaling
# In flipping, script goes through edit mode of objects
# Scale and rotate are apllied to all objects participating
#
# WISH LIST
#
# Abbility to include lattices and other non-mesh types in the transformation
# Abbility to adjust the rotation of the cage to better suite the geometry of scaled object
# Bgl overlay for mode points and eventualy the mode cage
# Blf instructions on screen, preferably interactive
# Cleaner code and faster performance
#
# WARNINGS
#
# The addon is still in beta and is not imune to hevy geometry, multiple modifiers, existing lattices and has not been tested in various scenarios, so avoid production use... 



bl_info={
  'name':'NP Cage Scale 016',
  'author':'Okavango with CoDEmanX, lukas_t, matali',
  'version':(0,1,6),
  'blender':(2,68,0),
  'location':'View3D',
  'description':'Scale and stretch objects using cage with handles - install, assign shortcut, save user settings',
  'category':'3D View'}

import bpy, math
import mathutils

# Defining the main class - the macro:
class NPCageScale016(bpy.types.Macro):
    bl_idname='object.np_cage_scale_016'
    bl_label='NP Cage Scale 016'
    bl_options={'REGISTER','UNDO'}

# Defining the storage class that will serve as a varable-bank for exchange among the classes. Later, this bank will recieve more variables with their values for safe keeping, as the program goes on:
class Storage:
    use_snap=None    
    snap_element=None
    snap_target=None

# Defining the first of the operational classes for aquiring the exact dimensions and the location of the selection - it reads the selection and scans all of it's vertices to find the most outer-ones in all directions. In this way it aquires the dimensions of the selection. The location of the selected group is calculated as the middle point afterwards: 
class NPCSGetSelDimLoc(bpy.types.Operator):
    bl_idname='object.np_cs_get_sel_dim_loc'
    bl_label='NP CS Get Sel Dim Loc'
    bl_options={'INTERNAL'}    

    def execute(self,context):
        # First, storing all of the system preferences set by the user, that will be changed during the process, in order to restore them when the operation is completed:
        Storage.use_snap=bpy.context.tool_settings.use_snap 
        Storage.snap_element=bpy.context.tool_settings.snap_element
        Storage.snap_target=bpy.context.tool_settings.snap_target
        Storage.pivot_point=bpy.context.space_data.pivot_point
        Storage.trans_orient=bpy.context.space_data.transform_orientation     
        # Reading and storing the selection:
        selob=bpy.context.selected_objects
        Storage.selob=selob
        # Applying previous transformations:
        bpy.ops.object.transform_apply(rotation=True,scale=True)
        # Preparing the environment for the operator functions:
        bpy.context.space_data.pivot_point='MEDIAN_POINT'
        bpy.context.space_data.transform_orientation='GLOBAL'         
        # Scanning the dimensions of the selection:
        fo=selob[0]
        foloc=fo.location
        xplus=foloc.x
        xminus=foloc.x
        yplus=foloc.y
        yminus=foloc.y
        zplus=foloc.z
        zminus=foloc.z
        for ob in selob:
            obdim=ob.dimensions
            foloc=ob.location
            obxplus=foloc.x+obdim.x/2
            obxminus=foloc.x-obdim.x/2 
            obyplus=foloc.y+obdim.y/2
            obyminus=foloc.y-obdim.y/2 
            obzplus=foloc.z+obdim.z/2
            obzminus=foloc.z-obdim.z/2                              
            if obxplus>xplus:
                xplus=obxplus
            if obxminus<xminus:
                xminus=obxminus
            if obyplus>yplus:
                yplus=obyplus
            if obyminus<yminus:
                yminus=obyminus
            if obzplus>zplus:
                zplus=obzplus
            if obzminus<zminus:
                zminus=obzminus
        # Calculating the dimensions and the location out of the outermost vertices:
        seldim=(xplus-xminus,yplus-yminus,zplus-zminus)
        Storage.seldim=seldim
        selloc=((xplus+xminus)/2,(yplus+yminus)/2,(zplus+zminus)/2)
        Storage.selloc=selloc
        return{'FINISHED'}

# Defining the class (operator) for adding a representation of the scale cage to the viewport. Upon this representation, the user will pick the type of transformation he wants to be performed:
class NPCSAddModeCage(bpy.types.Operator):
    bl_idname='object.np_cs_add_mode_cage'
    bl_label='NP CS Add Mode Cage'
    bl_options={'INTERNAL'}    
    
    def execute(self,context):
        selob=Storage.selob
        seldim=Storage.seldim
        selloc=Storage.selloc
        mcverts=(
        (-seldim[0]/2,-seldim[1]/2,-seldim[2]/2),
        (-seldim[0]/2,seldim[1]/2,-seldim[2]/2),
        (seldim[0]/2,seldim[1]/2,-seldim[2]/2),
        (seldim[0]/2,-seldim[1]/2,-seldim[2]/2),
        (-seldim[0]/2,-seldim[1]/2,seldim[2]/2),
        (-seldim[0]/2,seldim[1]/2,seldim[2]/2),
        (seldim[0]/2,seldim[1]/2,seldim[2]/2),
        (seldim[0]/2,-seldim[1]/2,seldim[2]/2),
        (seldim[0]/2,0,0),
        (-seldim[0]/2,0,0),
        (0,seldim[1]/2,0),
        (0,-seldim[1]/2,0),
        (0,0,seldim[2]/2),
        (0,0,-seldim[2]/2))
        mcedges=(
        (0,1),(0,3),(0,4),
        (1,2),(1,5),
        (2,3),(2,6),
        (3,7),
        (4,5),(4,7),
        (5,6),
        (6,7),
        (8,9),(10,11),(12,13))
        mcfaces=[]
        mcme=bpy.data.meshes.new('NP_CS_mode_cage')
        mcme.from_pydata(mcverts,mcedges,mcfaces)
        mcob=bpy.data.objects.new('NP_CS_mode_cage',mcme)
        mcob.location=selloc
        Storage.mcob=mcob
        scn=bpy.context.scene
        scn.objects.link(mcob)
        scn.objects.active=mcob
        scn.update()
        # Isolating the mode_cage for easier point selecting:
        bpy.ops.object.select_all(action='DESELECT')
        mcob.select=True
        bpy.ops.object.hide_view_set(unselected=True)
        return{'FINISHED'}
    
# Defining the operator for adding a dummy object on the original point of cursor location. 3D cursor will leave this position, go to the pointer, serve as an anchor generation location and come back, so there needs to be an object to safeguard his original position. After that, this dummy gets deleted:
class NPCSAddCursorDummy(bpy.types.Operator):
    bl_idname='object.np_cs_add_cursor_dummy'
    bl_label='NP CS Add Cursor Dummy'
    bl_options={'INTERNAL'}    
    
    def execute(self,context):
        bpy.ops.mesh.primitive_cube_add()
        cdummy=bpy.context.object
        cdummy.name='NP_CS_cdummy'
        Storage.cdummy=cdummy
        #print('020')         
        return{'FINISHED'}

# Defining the operator that will generate a one-vertex mesh i call anchor, at the spot marked by 3d cursor:
class NPCSAddAnchor(bpy.types.Operator):
    bl_idname='object.np_cs_add_anchor'
    bl_label='NP CS Add Anchor'
    bl_options={'INTERNAL'}    
    
    def execute(self, context):
        bpy.ops.mesh.primitive_cube_add(enter_editmode=True)
        bpy.ops.mesh.select_all
        bpy.ops.mesh.delete(type='VERT')
        bpy.ops.object.mode_set(mode='OBJECT')
        anchor=bpy.context.object
        anchor.name='NP_CS_anchor'
        Storage.anchor=anchor
        #print('030') 
        return{'FINISHED'}

# Activating dummy object for 3D cursor's return home:
class NPCSActivateCursorDummy(bpy.types.Operator):
    bl_idname='object.np_cs_activate_cursor_dummy'
    bl_label='NP CS Activate Cursor Dummy'
    bl_options={'INTERNAL'}    
    
    def execute(self,context):
        cdummy=Storage.cdummy
        bpy.context.scene.objects.active=cdummy
        Storage.cdummyloc=cdummy.location
        #print('040')
        return{'FINISHED'}

# Deleting dummy object and activating anchor for it's use in the pick-mode process:
class NPCSActivateAnchor(bpy.types.Operator):
    bl_idname='object.np_cs_activate_anchor'
    bl_label='NP CS Activate Anchor'
    bl_options={'INTERNAL'}    
    
    def execute(self,context):
        cdummy=Storage.cdummy
        bpy.ops.object.select_all(action='DESELECT')
        cdummy.select=True
        bpy.ops.object.delete('EXEC_DEFAULT')
        anchor=Storage.anchor
        bpy.context.scene.objects.active=anchor
        anchor.select=True
        # Preparing for the move operator, that will enable us to carry the anchor to desired point on the mode_cage. For this, we need to enable the specific snap parameters:
        bpy.context.tool_settings.use_snap=True
        bpy.context.tool_settings.snap_element='VERTEX'
        bpy.context.tool_settings.snap_target='ACTIVE'
        #print('050')
        return{'FINISHED'}

class NPCSRunModeSelect(bpy.types.Operator):
    bl_idname='object.np_cs_run_mode_select'
    bl_label='NP CS Run Mode Select'
    bl_options={'REGISTER','UNDO','INTERNAL'}

    count=0
       
    def modal(self,context,event):
        self.count+=1
        mcob=Storage.mcob
        selob=Storage.selob
        #print('080')
        if self.count == 1:
            bpy.ops.transform.translate('INVOKE_DEFAULT')
        elif event.type in ('LEFTMOUSE','RET','NUMPAD_ENTER','SPACE') and event.value=='RELEASE':
            return{'FINISHED'}
        elif event.type in ('ESC','RIGHTMOUSE'):
            mcob.select=True
            bpy.ops.object.delete('EXEC_DEFAULT') 
            bpy.ops.object.hide_view_clear()
            bpy.ops.object.select_all(action='DESELECT')
            for ob in selob:
                ob.select=True
            bpy.context.tool_settings.use_snap=Storage.use_snap
            bpy.context.tool_settings.snap_element=Storage.snap_element
            bpy.context.tool_settings.snap_target=Storage.snap_target
            bpy.context.space_data.pivot_point=Storage.pivot_point
            bpy.context.space_data.transform_orientation=Storage.trans_orient
            return{'CANCELLED'}
        return{'PASS_THROUGH'}

    def invoke(self,context,event):
        #print("START_____")
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}    
    
    
# Defining the operator that will read user input on the mode cage and turn it in the appropriate mode for transformation:
class NPCSReadMode(bpy.types.Operator):
    bl_idname='object.np_cs_read_mode'
    bl_label='NP CS Read Mode'
    bl_options={'INTERNAL'}    
    
    def execute(self,context):
        mcob=Storage.mcob
        mcme=mcob.data
        anchor=Storage.anchor
        anchorloc=anchor.location
        selloc=Storage.selloc
        selloc=mathutils.Vector(selloc)
        mcvertmin=(1000,1000,1000)
        mcvertmin=mathutils.Vector(mcvertmin)
        mcvertmind=mcvertmin.length
        mode=0
        for n,mcvert in enumerate(mcme.vertices):
            delta=mcvert.co+selloc-anchorloc
            deltad=delta.length
            if deltad<mcvertmind:
                mcvertmind=deltad
                mode=n
        seldim=Storage.seldim
        seldimmin=seldim[0]
        for s in seldim:
            if s<seldimmin:
                seldimmin=s
        margin=seldimmin/10
        if mcvertmind>margin:
            mode=-1
        print('mode',mode)
        Storage.mode=mode
        mcob.select=True
        bpy.ops.object.delete('EXEC_DEFAULT')
        # returning to the normal view:
        bpy.context.tool_settings.use_snap= Storage.use_snap
        bpy.ops.object.hide_view_clear()
        bpy.ops.object.select_all(action='DESELECT')
        #print('060')
        return {'FINISHED'}
    
# Defining the operator that will take the mode information and create the appropriate transform cage. Then it will prepare it for the actual transformation:
class NPCSAddTransCage(bpy.types.Operator):
    bl_idname='object.np_cs_add_trans_cage'
    bl_label='NP CS Add Trans Cage'
    bl_options={'INTERNAL'}    
    
    def execute(self,context):
        seldim=Storage.seldim
        selloc=Storage.selloc
        selob=Storage.selob
        cdummyloc=Storage.cdummyloc
        mode=Storage.mode
        #print('persportho',bpy.context.space_data.region_3d.view_perspective)
        if mode>-2 and mode<8:
            bpy.ops.mesh.primitive_cube_add(view_align=False, enter_editmode=False,location=selloc,rotation=(0.0, 0.0, 0.0))
            transcube=bpy.context.object
            transcube.name='NP_CS_transcube'
            transcube.scale=(0.5,0.5,0.5)
            bpy.ops.object.select_all(action='DESELECT')
            transcube.select=True
            bpy.ops.object.transform_apply(rotation=True,scale=True)
            transcube.draw_type='WIRE'
            transcube.scale=seldim
            bpy.ops.object.transform_apply(rotation=True,scale=True)
            Storage.transcube=transcube
            tcme=transcube.data
            selloc=mathutils.Vector(selloc)
            if mode==-1:
                v=selloc
            elif mode==0:
                v=tcme.vertices[6].co+selloc 
            elif mode==1:
                v=tcme.vertices[7].co+selloc 
            elif mode==2:
                v=tcme.vertices[4].co+selloc          
            elif mode==3:
                v=tcme.vertices[5].co+selloc 
            elif mode==4:
                v=tcme.vertices[2].co+selloc 
            elif mode==5:
                v=tcme.vertices[3].co+selloc 
            elif mode==6:
                v=tcme.vertices[0].co+selloc 
            elif mode==7:
                v=tcme.vertices[1].co+selloc    
            v=v.to_tuple(4)
            bpy.ops.object.add(type='EMPTY')
            c3dfirst=bpy.context.object
            bpy.ops.object.add(type='EMPTY',location=v)
            c3dsecond=bpy.context.object           
            bpy.ops.view3d.snap_cursor_to_active()
            bpy.ops.object.select_all(action='DESELECT')                
            for ob in selob:
                ob.select=True
            transcube.select=True
            bpy.ops.object.origin_set(type='ORIGIN_CURSOR')
            bpy.context.scene.objects.active=c3dfirst
            bpy.ops.view3d.snap_cursor_to_active()                
            bpy.ops.object.select_all(action='DESELECT')
            c3dfirst.select=True
            c3dsecond.select=True
            bpy.ops.object.delete('EXEC_DEFAULT')
            for ob in selob:
                ob.select=True
            transcube.select=True                              
        else:
            bpy.ops.object.add(type='LATTICE',view_align=False, enter_editmode=False,location=selloc,rotation=(0.0, 0.0, 0.0))
            translatt=bpy.context.object
            translatt.name='NP_CS_translattice'
            bpy.ops.object.select_all(action='DESELECT')
            translatt.select=True
            bpy.ops.object.transform_apply(rotation=True,scale=True)
            translatt.draw_type='WIRE'
            Storage.translatt=translatt
            tlme=translatt.data
            eul90=math.radians(90) #i just can't believe i have to do this 
            eul180=math.radians(180)
            eul270=math.radians(270)
            bpy.types.SpaceView3D.transform_orientation='GLOBAL'
            if mode==8:
                translatt.rotation_euler=(0,eul90,0)                       
            elif mode==9:
                translatt.rotation_euler=(0,eul270,0)                      
            elif mode==10:
                translatt.rotation_euler=(eul270,0,0)
            elif mode==11:
                translatt.rotation_euler=(eul90,0,0)      
            elif mode==12:
                translatt.rotation_euler=(0,0,0)
            elif mode==13:
                translatt.rotation_euler=(eul180,0,0)
            bpy.types.SpaceView3D.transform_orientation='LOCAL'
            bpy.context.scene.objects.active=translatt
            bpy.ops.transform.resize(value=seldim, constraint_orientation='LOCAL')
            tlme.interpolation_type_u='KEY_LINEAR'
            tlme.interpolation_type_v='KEY_LINEAR'
            tlme.interpolation_type_w='KEY_LINEAR'
            sellen=len(selob)
            print('sellen',sellen)
            for ob in selob:
                bpy.context.scene.objects.active=ob
                bpy.ops.object.modifier_add(type='LATTICE')
                modlen=len(ob.modifiers)
                modmod=ob.modifiers[modlen-1]
                modmod.name='NP_lat_mod_'+str(modlen-1)
                #print('modmod',modmod)
                #print('modmod.name',modmod.name)
                modmod.object=translatt
            bpy.context.scene.objects.active=translatt
            bpy.ops.object.mode_set(mode='EDIT')                           
            for n,pt in enumerate(tlme.points):
                    if n==4 or n==5 or n==6 or n==7:          
                        pt.select=True  
            bpy.context.tool_settings.use_snap = False
            bpy.context.tool_settings.snap_element='VERTEX'
            bpy.context.tool_settings.snap_target='ACTIVE'
        #print('070')
        #print(Storage.toggle)     
        return {'FINISHED'}

# Defining the operator that will perform the transformation based on the chosen mode. This is the heart of this addon and it now has the abblity to choose between two diferent modal operators. It also uses some listening operators that clean up the leftovers should the user interrupt the command. Many thanks to CoDEmanX and lukas_t:
class NPCSRunTransform(bpy.types.Operator):
    bl_idname='object.np_cs_run_transform'
    bl_label='NP CS Run Transform'
    bl_options={'REGISTER','UNDO','INTERNAL'}
    
    count=0
       
    def modal(self,context,event):
        mode=Storage.mode
        self.count+=1
        #print('080')
        if self.count == 1:
            if mode>-2 and mode<8:
                bpy.ops.transform.resize('INVOKE_DEFAULT')
            else:
                bpy.ops.transform.translate('INVOKE_DEFAULT',constraint_axis=(False,False,True),constraint_orientation='LOCAL')
        elif event.type in ('LEFTMOUSE','RET','NUMPAD_ENTER','SPACE') and event.value=='RELEASE':
            bpy.ops.object.np_cs_delete_trans_cage()
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')
            return{'FINISHED'}
        elif event.type in ('ESC','RIGHTMOUSE'):
            bpy.ops.object.np_cs_delete_trans_cage()
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')            
            return{'FINISHED'}
        return{'PASS_THROUGH'}

    def invoke(self,context,event):
        #print("START_____")
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

# Defining the operator that will delete the transcage after it's use:
class NPCSDeleteTransCage(bpy.types.Operator):
    bl_idname='object.np_cs_delete_trans_cage'
    bl_label='NP CS Delete Trans Cage'
    bl_options = {'REGISTER','UNDO','INTERNAL'}
       
    def execute(self,context):
        mode=Storage.mode
        selob=Storage.selob
        #print('090')
        if mode>-2 and mode<8:
            #bpy.ops.object.mode_set(mode='OBJECT')           
            bpy.ops.object.select_all(action='DESELECT')
            transcube=Storage.transcube
            transcube.select=True
            bpy.ops.object.delete('EXEC_DEFAULT')
            for ob in selob:
                ob.select=True
        else:
            bpy.ops.object.mode_set(mode='OBJECT')
            translatt=Storage.translatt
            for ob in selob:
                ob.select=True
                bpy.context.scene.objects.active=ob
                modlen=len(ob.modifiers)
                modmod=ob.modifiers[modlen-1]
                #print('modmod',modmod)
                #print('modmod.name',modmod.name)
                bpy.ops.object.modifier_apply(apply_as='DATA', modifier=modmod.name)
                bpy.ops.object.modifier_remove(modifier=modmod.name)
                ob.select=False
            bpy.ops.object.select_all(action='DESELECT')
            translatt.select=True
            bpy.ops.object.delete('EXEC_DEFAULT')
            for ob in selob:
                ob.select=True
        bpy.context.tool_settings.use_snap=Storage.use_snap
        bpy.context.tool_settings.snap_element=Storage.snap_element
        bpy.context.tool_settings.snap_target=Storage.snap_target
        bpy.context.space_data.pivot_point=Storage.pivot_point
        bpy.context.space_data.transform_orientation=Storage.trans_orient
        return {'FINISHED'}
    
# This is the actual addon process, the algorithm that defines the order of operator activation inside the main macro:           
def register():
    bpy.utils.register_module(__name__)
    
    NPCageScale016.define("OBJECT_OT_np_cs_get_sel_dim_loc")
    NPCageScale016.define("OBJECT_OT_np_cs_add_mode_cage")
    NPCageScale016.define("OBJECT_OT_np_cs_add_cursor_dummy")
    NPCageScale016.define("VIEW3D_OT_cursor3d")
    NPCageScale016.define("OBJECT_OT_np_cs_add_anchor")
    NPCageScale016.define("OBJECT_OT_np_cs_activate_cursor_dummy")
    NPCageScale016.define("VIEW3D_OT_snap_cursor_to_active")
    NPCageScale016.define("OBJECT_OT_np_cs_activate_anchor")
    NPCageScale016.define("OBJECT_OT_np_cs_run_mode_select")
    NPCageScale016.define("OBJECT_OT_np_cs_read_mode")
    NPCageScale016.define("OBJECT_OT_np_cs_add_trans_cage")
    NPCageScale016.define("OBJECT_OT_np_cs_run_transform")
    #NPCageScale012.define("OBJECT_OT_np_cs_delete_trans_cage")
    #m=NPCageScale012.define("OBJECT_OT_origin_set")
    #m.properties.type='ORIGIN_GEOMETRY'

def unregister():
  bpy.utils.unregister_module(__name__)
       
if __name__ == "__main__":
  register()

